/*     
 * $RCSfile: upi.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: upi.h,v $
 * Revision 1.84.1.3  1994/05/05  23:00:22  menze
 * Added _CTL definitions for protocols in the psync suite
 *
 * Revision 1.84.1.2  1994/04/19  00:22:42  menze
 * Protocol ID field is now ProtId type
 *
 * Revision 1.84.1.1  1994/03/17  22:20:34  menze
 * Support for FDDIhost type
 *
 * Revision 1.84  1994/03/11  21:19:15  dcs
 * Added control numbers for kerb_services, kerb_serv_bottom, and
 * recvauth protocols.
 *
 * Revision 1.83  1993/12/16  23:21:10  menze
 * Added MAC_CTL and FDDI_CTL
 *
 * Revision 1.82  1993/12/11  00:23:25  menze
 * fixed #endif comments
 *
 * Revision 1.81  1993/11/15  20:49:48  menze
 * Added SIM_CTL
 *
 */

#ifndef upi_h
#define upi_h

#include "xtype.h"
#include "idmap.h"
#include "msg_s.h"
#include "part.h"
#include "platform.h"

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE  0
#endif


typedef enum {Protocol,Session} XObjType;

/* default number of down protocols or sessions */
#define STD_DOWN 8


#ifdef __STDC__

typedef	struct xobj *	(* XOpenFunc)
    (struct xobj *, struct xobj *, struct xobj *, Part * );
typedef	xkern_return_t	(* XCloseFunc) (struct xobj * );
typedef	xkern_return_t	(* XCloseDoneFunc) ( struct xobj * );
typedef	xkern_return_t	(* XOpenEnableFunc)
    ( struct xobj *, struct xobj *, struct xobj *, Part * );
typedef	xkern_return_t	(* XOpenDisableFunc)
    ( struct xobj *, struct xobj *, struct xobj *, Part * );
typedef	xkern_return_t	(* XOpenDisableAllFunc)
    ( struct xobj *, struct xobj *);
typedef	xkern_return_t	(* XOpenDoneFunc)
    ( struct xobj *, struct xobj *, struct xobj *, struct xobj * );
typedef	xkern_return_t	(* XDemuxFunc)
    ( struct xobj *, struct xobj *, Msg * );
typedef	xkern_return_t	(* XCallDemuxFunc)
    ( struct xobj *, struct xobj *, Msg *, Msg * );
typedef	xkern_return_t	(* XPopFunc)
    ( struct xobj *, struct xobj *, Msg *, void * );
typedef	xkern_return_t	(* XCallPopFunc)
    ( struct xobj *, struct xobj *, Msg *, void *, Msg * );
typedef	xmsg_handle_t	(* XPushFunc) ( struct xobj *, Msg * );
typedef	xkern_return_t	(* XCallFunc) ( struct xobj *, Msg *, Msg * );
typedef int		(* XControlFunc)( struct xobj *, int, char *, int );
typedef	xkern_return_t	(* XDuplicateFunc) ( struct xobj * );
typedef	xkern_return_t	(* XShutDownFunc) ( struct xobj * );

#else

typedef  Pfo	XOpenFunc;
typedef  Pfk 	XCloseFunc;
typedef  Pfk    XCloseDoneFunc;
typedef  Pfk	XOpenEnableFunc;
typedef  Pfk	XOpenDisableFunc;
typedef  Pfk	XOpenDisableAllFunc;
typedef  Pfk	XOpenDoneFunc;
typedef  Pfk 	XDemuxFunc;
typedef  Pfk 	XCallDemuxFunc;
typedef  Pfk 	XPopFunc;
typedef  Pfk 	XCallPopFunc;
typedef  Pfh 	XPushFunc;
typedef  Pfk 	XCallFunc;
typedef  Pfi 	XControlFunc; 
typedef  Pfk	XDuplicateFunc;
typedef  Pfk	XShutDownFunc;

#endif

typedef	struct xobj {
  XObjType type;
  char  *name;
  char  *instName;
  char  *fullName;
  VOID	*state;
  Bind  binding;
  int   rcnt;
  xk_int32	id;
  int	*traceVar;
  XOpenFunc open;             
  XCloseFunc close;            
  XCloseDoneFunc closedone;             
  XOpenEnableFunc openenable;       
  XOpenDisableFunc opendisable;      
  XOpenDisableAllFunc opendisableall;   
  XOpenDoneFunc opendone;         
  XDemuxFunc demux;            
  XCallDemuxFunc calldemux;        
  XPopFunc pop;              
  XCallPopFunc callpop;          
  XPushFunc push;             
  XCallFunc call;             
  XControlFunc control;          
  XDuplicateFunc duplicate;        
  XShutDownFunc shutdown;         
  int 	numdown;
  int	downlistsz;
  unsigned char	idle;
  struct xobj  *down[STD_DOWN];
  struct xobj **downlist;
  struct xobj  *myprotl;
  struct xobj  *up;
  struct xobj  *hlpType;
} *XObj;


typedef struct xenable {
  XObj	hlpRcv;
  XObj	hlpType;
  Bind	binding;
  int	rcnt;
} Enable;


typedef XObj Sessn;
typedef XObj Protl;

extern int globalArgc;
extern char **globalArgv;

#include "msg.h"



/* error stuff */


#define ERR_XK_MSG	((xmsg_handle_t) XK_FAILURE)
#define	ERR_XOBJ	((XObj) XK_FAILURE)
#define	ERR_SESSN	((Sessn) XK_FAILURE)
#define	ERR_PROTL	((Protl) XK_FAILURE)
#define ERR_ENABLE	((Enable *) XK_FAILURE)
#define ERR_XMALLOC	0

/* protocol and session operations */

#ifdef __STDC__

extern	XObj xOpen( XObj hlpRcv, XObj hlpType, XObj llp, Part *p );
extern	xkern_return_t
  xOpenEnable( XObj hlpRcv, XObj hlpType, XObj llp, Part *p );
extern	xkern_return_t
  xOpenDisable( XObj hlpRcv, XObj hlpType, XObj llp, Part *p );
extern	xkern_return_t  xOpenDisableAll( XObj hlpRcv, XObj llp );
extern	xkern_return_t  xOpenDone( XObj hlp, XObj s, XObj llp );
extern	xkern_return_t  xCloseDone( XObj s );
extern	xkern_return_t  xClose( XObj s );
extern	xkern_return_t  xDemux( XObj s, Msg *msg );
extern	xkern_return_t  xCallDemux( XObj s, Msg *msg, Msg *returnmsg );
extern	xmsg_handle_t	xPush( XObj s, Msg *msg );
extern	xkern_return_t  xCall( XObj s, Msg *msg, Msg *returnmsg );
extern  xkern_return_t	xShutDown( XObj );

/* 
 * xPop and xCallPop prototypes are in upi_inline.h
 */

extern	int  		xControl( XObj s, int opcode, char *buf, int len );
extern  xkern_return_t  xDuplicate( XObj s );

typedef void	(* DisableAllFunc)( VOID *, Enable * );

xkern_return_t	defaultOpenDisable( Map, XObj, XObj, VOID * );
xkern_return_t	defaultOpenDisableAll( Map, XObj, DisableAllFunc);
xkern_return_t	defaultOpenEnable( Map, XObj, XObj, VOID * );
xkern_return_t	defaultVirtualOpenDisable( XObj, Map, XObj, XObj, XObj *,
					   Part *);
xkern_return_t	defaultVirtualOpenEnable( XObj, Map, XObj, XObj, XObj *,
					  Part *);

#else

extern	XObj xOpen();
extern	xkern_return_t  xOpenEnable();
extern	xkern_return_t  xOpenDisable();
extern	xkern_return_t  xOpenDone();
extern	xkern_return_t  xCloseDone();
extern	xkern_return_t  xClose();
extern	xkern_return_t  xDemux();
extern	xkern_return_t  xCallDemux();
extern	xmsg_handle_t 	xPush();
extern	xkern_return_t  xCall();
extern	int  		xControl();
extern  xkern_return_t  xDuplicate();
extern  xkern_return_t	xShutDown();

typedef void	(* DisableAllFunc)();

xkern_return_t	defaultOpenDisable();
xkern_return_t	defaultOpenEnable();
xkern_return_t	defaultVirtualOpenDisable();
xkern_return_t	defaultVirtualOpenEnable();

#endif

/* initialization operations */

#ifdef __STDC__

typedef void	(*Pfvo)( XObj );
typedef Pfvo XObjInitFunc;

extern	XObj xCreateSessn(XObjInitFunc f, XObj hlpRcv, XObj hlpType, XObj llp,
			  int downc, XObj *downv);
extern	XObj xCreateProtl(XObjInitFunc f, char *nm, char *inst, int *trace,
			  int downc, XObj *downv);
extern	xkern_return_t  xDestroy(XObj s);
extern  void		upiInit( void );

#else

typedef void	(*Pfvo)();
typedef Pfvo XObjInitFunc;

extern XObj 		xCreateSessn();
extern XObj 		xCreateProtl();
extern xkern_return_t	xDestroy();
extern void		upiInit();

#endif

/* utility routines */

#ifdef __STDC__

extern	XObj 	xGetProtlByName(char *name);
extern  bool	xIsValidXObj( XObj );
extern	xkern_return_t  xSetDown(XObj self, int i, XObj object);
extern	XObj 	xGetDown(XObj self, int i);
extern	void	xPrintXObj( XObj );
extern  void	xRapture( void );

#else

extern	XObj 	xGetProtlByName();
extern  bool	xIsValidXObj();
extern	xkern_return_t  xSetDown();
extern	XObj 	xGetDown();
extern	void	xPrintXObj();
extern  void	xRapture();

#endif /* __STDC__ */


/* object macros */

#define xSetUp(s, hlp)	((s)->up = (hlp))
#define xGetUp(s)	((s)->up)
#define xMyProtl(s) ((s)->myprotl)
#define xIsXObj(s) ((s) && (s) != ERR_XOBJ)
#define xIsSession(s) ((s) && (s) != ERR_XOBJ && (s)->type == Session)
#define xIsProtocol(s) ((s) && (s) != ERR_XOBJ && (s)->type == Protocol)
#define xHlpType(s)	((s)->hlpType)

/*
 * control operation definitions
 *
 * NOTE: if you change the standard control ops, make the
 * corresponding change to the controlops string array in upi.c
 */
enum {
    GETMYHOST = 0,	/* standard control operations        */
    GETMYHOSTCOUNT,	/* common to all protocols            */
    GETPEERHOST,
    GETPEERHOSTCOUNT,
    GETBCASTHOST,
    GETMAXPACKET,
    GETOPTPACKET,
    GETMYPROTO,
    GETPEERPROTO,
    RESOLVE,
    RRESOLVE,
    FREERESOURCES,
    GETPARTICIPANTS,
    SETNONBLOCKINGIO
};

#define	ARP_CTL		1	/* like a protocol number; used to    */
#define	BLAST_CTL	2	/* partition opcode space	      */
#define	ETH_CTL		3
#define	IP_CTL		4
#define	SCR_CTL		5
#define	VCHAN_CTL	6
#define	PSYNC_CTL	7
#define	SS_CTL		8
#define	SUNRPC_CTL	9
#define	NFS_CTL		10
#define	TCP_CTL		11
#define UDP_CTL		12
#define ICMP_CTL	13
#define VNET_CTL	14
#define BIDCTL_CTL	15
#define CHAN_CTL	16
#define VDROP_CTL	17
#define KM_CTL		18
#define VTAP_CTL	19
#define JOIN_CTL	20
#define CRYPT_CTL	21
#define BIND_CTL	22
#define SIMETH_CTL	23
#define HASH_CTL	24
#define SIM_CTL		25
#define MAC_CTL		26
#define FDDI_CTL	27
#define KERBSERV_CTL	28
#define KERBSERVBOT_CTL 29
#define RECVAUTH_CTL    30
#define TOTAL_CTL	31
#define DISPATCH_CTL	32
#define DIVIDER_CTL	33

#define	TMP0_CTL	100	/* for use by new/tmp protocols until */
#define	TMP1_CTL	101	/* a standard CTL number is assigned  */
#define	TMP2_CTL	102
#define	TMP3_CTL	103
#define	TMP4_CTL	104

#define	MAXOPS		100	/* maximum number of ops per protocol */

/* Check the length of a control argument */
#define checkLen(A, B) { \
  if ((A) < (B)) { \
    return -1; \
  } \
}

#include "ip_host.h"
#include "eth_host.h"
#include "fddi_host.h"

#ifdef __STDC__

extern xkern_return_t	str2ipHost( IPhost *, char * );
extern xkern_return_t	str2ethHost( ETHhost *, char * );
extern xkern_return_t	str2fddiHost( FDDIhost *, char * );
extern char *	ipHostStr( IPhost * );
extern char *	ethHostStr( ETHhost * );
extern char *	fddiHostStr( FDDIhost * );

#else

extern xkern_return_t	str2ipHost();
extern xkern_return_t	str2ethHost();
extern xkern_return_t	str2fddiHost();
extern char *	ipHostStr();
extern char *	ethHostStr();
extern char *	fddiHostStr();

#endif /* __STDC__ */


extern XObj	xNullProtl;


#define X_NEW(Type) (Type *)xMalloc(sizeof(Type))
#define FREE(X) (free(X), (X) = 0)

/*
 * Optimize xDemux, xCallDemux, xPush and xCall as macros.  The other critical
 * path upi functions (xPop and xCallPop) are defined as inline functions
 * in upi_inline.h
 */
#if ! defined(XK_DEBUG) || defined(XK_UPI_MACROS)

#   define xDemux(DS, M) ((*((DS)->up->demux))((DS)->up, (DS), (M)))
#   define xCallDemux(DS, M, RM) ((*((DS)->up->calldemux))((DS)->up, (DS), (M), (RM)))
#   define xPush(S, M) ((*((S)->push))((S), (M)))
#   define xCall(S, M, RM) ((*((S)->call))((S), (M), (RM)))

#endif /* ! XK_DEBUG */


#endif

