/*
 * $RCSfile: sartest.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1991  Arizona Board of Regents
 *
 *
 * $Log: sartest.c,v $
 * Revision 1.3.1.1  1994/01/24  17:50:08  menze
 * Fixed #endif comments
 *
 */

/*#define CSUM*/

/*
 * Ping-pong test of the SAR protocol
 */

#include "site.h"
#include "eth.h"
#include "sarProt.h"
#include "msg.h"

#ifdef __STDC__
void showStatistics( XOBj );
#else
void showStatistics();
#endif

#define HOST_TYPE SARhost
#define INIT_FUNC sartest_init
#define TRACE_VAR sartestp
#define PROT_STRING "sar"

static HOST_TYPE ServerAddr = { 0, 0, 0};
static HOST_TYPE ClientAddr = {0x800,0x2b1c,0x18e8};

#define TRIPS 10000 /*00*/
#define TIMES 1
#define DELAY 3

/*#define ACK*/
#define PP
#define CORRECT_ERRORS

/*
 * Define to do timing calculations
 */
#define TIME


static int lens[] = { 
/*  44, 1024, 2048, 4096, 8192, 16384, 32768, 65536*/
  44, 1012, 2024, 4004, 8008, 16016, 32032, 64064
};


#define SAVE_SERVER_SESSN
#define CLIENT_OPENENABLE

static MNodeAllocator *myAllocator;

#include "common_test.c"

#undef TIME


static int word;
static int rcvdMsgs;
static int bufs;
static int currentLen;
static int msgs;
static Semaphore sem;
static int lastMsgs;

void
  sarWatchdog(ev, arg)
Event *ev;
VOID *arg;
{
  XObj self = (XObj)arg;

  if (lastMsgs == rcvdMsgs) {
    printf("Watchdog barking\n");
    showStatistics(self);
    semSignal(&sem);
  }
  lastMsgs = rcvdMsgs;
  evDetach(evSchedule(sarWatchdog, self, 20000000));
}

static int
sarIsServer(self)
     XObj self;
{
  return bcmp((char *)&myHost, (char *)&ClientAddr, sizeof(HOST_TYPE));
}

static int
sarIsClient(self)
     XObj self;
{
  int result;

  printf("myHost=%x:%x:%x,ClientAddr=%x:%x:%x\n",
	 myHost.high, myHost.mid,myHost.low,
	 ClientAddr.high,ClientAddr.mid,ClientAddr.low);

  result = ! bcmp((char *)&myHost, (char *)&ClientAddr, sizeof(HOST_TYPE));

#ifdef ACK
  if (result) {
    semInit(&sem, 3); /* window size = 4 * 16 buffers */
    evDetach(evSchedule(sarWatchdog, self, 60000000));
    Delay(5000);
  }
#endif /* ACK */

#ifdef PP
  if (result) {
    semInit(&sem, 0);
    evDetach(evSchedule(sarWatchdog, self, 60000000));
    Delay(5000);
  }
#endif

  return result;
}

void
  showStatistics(self)
    XObj self;
{
  int stats[11];
  
  xControl(xGetDown(self, 0), SAR_GETSTATISTICS, (char*)stats, 
	   sizeof(stats));

  printf("xmitIdles=%d, xmitFull=%d, xmitEmpty=%d\n", 
	 stats[0], stats[1], stats[2]);
  printf("rcvCRCErrors=%d, ", stats[3]);
  printf("rcvBQueueUnderflows=%d, ", stats[4]);
  printf("rcvQueueOverflows=%d, ", stats[5]);
  printf("rcvOutOfOrderCells=%d,\n", stats[6]);
  printf("rcvOutOfOrderFrags=%d, ", stats[7]);
  printf("rcvIdles=%d, ", stats[8]);
  printf("rcvHdrFIFOFull=%d, ", stats[9]);
  printf("rcvDMAFIFOFull=%d\n", stats[10]);
}


static bool 
show(ptr, len, arg)
char *ptr;
int len;
char *arg;
{
  int i, isav, j, *iptr = (int*)ptr;
  int start, *startp;
  int error = 0;

  /*printf("addr=%x, len=%d\n", ptr, len);*/
  
#ifndef TIME
  /*word = ((word + 10) / 11) * 11;*/

  for (i = 0; i < len / 4; i++, iptr++, word++) {  

    if ((word % 11) == 0) {
      /* new cell */
      if (error) printf("\n");
      error = 0;
      isav = i;
      start = word;
      startp = iptr;
    }

    if (*iptr != (((word / 11) << 16) | (word % 11)) ) {
      /*
       * verification error
       */
      if (*iptr == (((word + 10) / 11) << 16) ) {
	/* 
	 * xmitter sent a padded cell
	 */
	word = ((word + 10) / 11) * 11;
      }
      else if (!error) {
	/*
	 * show the entire cell
	 */
	i = isav;
	word = start;
	iptr = startp;

	printf("\nmsg %d, cell %d, pa=%x: ", rcvdMsgs, word / 11,
	       kvtophys((unsigned)iptr & ~(4096-1)) | 
	       ((unsigned)iptr & (4096-1)) );
	error = 1;
      }
    }

    if (error) {
      if (*iptr == (((word / 11) << 16) | (word % 11)))
	printf(" ok ");
      else  {
	printf(" %d,%d ", (*iptr >> 16) & 0xffff, *iptr & 0xffff); 
#ifdef CORRECT_ERRORS
	*iptr = (((word / 11) << 16) | ( word % 11));
#endif
      }
    }

  }
#endif /* TIME */

  return TRUE;
}

static xkern_return_t
sarServerDemux( self, lls, dg )
    XObj self, lls;
    Msg *dg;
{
  int i, len = msgLen(dg);

  rcvdMsgs++;
  /*printf("s%d ", rcvdMsgs);*/

  word = 0;
  msgForEach(dg, show, NULL);

#ifdef CSUM
  inCkSum(dg, &word, 4);
#endif

#ifdef ACK
  if (++msgs == TRIPS) {
    printf("\nsarServerDemux, received %d messages with length=%d\n", 
	   msgs, len);
    showStatistics(self);
    currentLen = len;
    msgs = 0;
  }

  bufs += (len + 4092) / 4092;
  if  (bufs >= 16) {
    bufs -= 16;

    msgTruncate(dg, 40);
    xPush(lls, dg);
  }
#endif /* ACK */

#ifdef PP
  xPush(lls, dg);
#endif /* PP */

  return XK_SUCCESS;
}


static xkern_return_t
sarClientDemux( self, lls, dg )
    XObj self, lls;
    Msg *dg;
{
  rcvdMsgs++;
  /*printf("c%d ", rcvdMsgs);*/

#ifdef PP
  word = 0;
  msgForEach(dg, show, NULL);
#endif

  semSignal(&sem);
  return XK_SUCCESS;
}



static int
sarRunTest( self, len, testNum )
    XObj self;
    int len;
    int testNum;
{
    Msg		savedMsg, msg;
    int 	test, trips, i;
    int         total = 0;
    char	*buf;
    XTime       startTime, endTime, netTime;
    unsigned    offset = 0;
    int *ptr;
    PState	*ps = (PState *)self->state;

    currentLen = len;
    rcvdMsgs = 0;

    lastMsgs = -1;
    self->demux = sarClientDemux;

#ifdef undef
    msgConstructAllocate(&savedMsg, len + 0x1000, &buf);
    /* page-align the buffer */
    offset = ((unsigned)(buf + 0x1000) & 0xfffff000) - (unsigned)buf;
    msgPopDiscard(&savedMsg, offset);
    msgTruncate(&savedMsg, len);
#endif

    msgConstructAllocateSetAllocator(&savedMsg, len, &buf, myAllocator);

#ifndef TIME    
    /* fill the buffer */
    ptr = (int*)(buf + offset);
    for (i = 0; ptr < (int*)(buf + offset + len); i++) {
      if ( ((unsigned)ptr & ~(0x1000 - 1)) == 0 && (i % 11) != 0 )
	i = ((i + 10) / 11) * 11;
      *ptr++ = ((i / 11) << 16) | (i % 11);
    }
#endif

    for (test = 0; test < TIMES; test++) {
	msgConstructCopy(&msg, &savedMsg);
	printf("Sending (%d) ...\n", ++total);
	/*msgForEach(&msg, show, NULL);*/
	printf("msg length: %d\n", msgLen(&msg));

	xGetTime(&startTime);

	for (trips = 0; trips < TRIPS; trips++) {
	  ps->clientPushResult = xPush(ps->lls, &msg);
	  if (ps->clientPushResult == ERR_XK_MSG) 
	    printf("sarRunTest: xPush faild, len=%d, test=%d, trip=%d\n",
		   len, test, trips);

#ifdef ACK
	  bufs += (len + 4092) / 4092;
	  if  (bufs >= 16) {
	    semWait(&sem);
	    bufs -= 16;
	  }
#endif
#ifdef PP
	  semWait(&sem);
#endif
	}

	xGetTime(&endTime);
	subtime(&startTime, &endTime, &netTime);
	printf("\nlen = %4d, %d trips: %6d.%06d\n", 
	       len, trips, netTime.sec, netTime.usec);
	msgDestroy(&msg);

	/* wait until all msgs are received */
	Delay(5000);
	printf("received %d messages\n", rcvdMsgs);
    }
    msgDestroy(&savedMsg); 
    showStatistics(self);
    return 0;
}

static void
sarServerSetPart( p )
    Part *p;
{
    partInit(p, 1);
    partPush(*p, 1, 4);
}

static void
testInit()
{
  myAllocator = newFbufAllocator(TRUE, TRUE);
  xAssert(myAllocator);

  isClient = sarIsClient;
  isServer = sarIsServer;
  runTest = sarRunTest;
  bufs = 0;
  currentLen = 0;
  rcvdMsgs = 0;
  msgs = 0;
  serverDemux = sarServerDemux;
}



