/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: xinit.c,v $
 *
 * HISTORY
 * $Log: xinit.c,v $
 * Revision 1.2  1994/06/11  00:02:25  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/04/22  02:08:20  davidm
 * Initial revision
 *
 */
/*
 * Initialize and start x-kernel.
 */
#include "scout_time.h"
#include "scout_boot.h"
#include "scout_config.h"
#include "xkernel.h"
#include "event_i.h"

int traceinit = 100;


int globalArgc;
char **globalArgv;


/*
 * Maximum number of args:
 */
#define NMAXARGS	16

/* initialization of testArgs determines default args: */
char *testArgs = 0;


static char *testArgv[NMAXARGS]	= {"xkernel", "-s"};
static int testArgc = 2;

/*
 * Split "testArgs" into individual arguments and store them in
 * testArgv[]; testArgc is the number of args in testArgv[];
 * testArgv[0] is always "xkernel".
 */
static void
split_args()
{
    extern char* index();
    char *cp = testArgs;

    /* split args and store them in testArgc/testArgv: */

    if (cp && *cp) {
	testArgc = 1;
    } /* if */
    while (cp && *cp) {
	testArgv[testArgc++] = cp;
	if (testArgc >= NMAXARGS) {
	    Kabort("init.c: too many actual args (increase NMAXARGS)");
	} /* if */
	cp = index(cp, ' ');
	if (cp) {
	    *cp++ = '\0';
	    while (*cp == ' ') {
		++cp;
	    } /* while */
	} /* if */
    } /* while */
    testArgv[testArgc] = 0;
} /* split_args */


static void
dump_config(void)
{
    printf("event manager: wheel size=%d, thread cache size=%d\n",
	   XK_EVENT_WHEEL_SIZE, XK_EVENT_THREAD_CACHE_SIZE);
    printf("shepherds: prolog="
#ifdef XK_SHEPHERD_PROLOG
	   "on"
#else
	   "off"
#endif
	   "; # of threads=%d; # of buffers=%d\n",
	   XK_SHEPHERD_THREADS, XK_SHEPHERD_BUFFERS);
    printf("misc: max frame size=%d; max link-header size=%d; "
	   "interleaved frames "
#ifdef XK_LANCE_INTERLEAVED
	   "on"
#else
	   "off"
#endif
	   "\n", XK_MAX_FRAME_SIZE, XK_MAX_HARDWARE_HEADER_SIZE);
} /* dump_config */


void
xkInit(void)
{
    int ticks_per_sec;

    split_args();
    globalArgc = testArgc;
    globalArgv = testArgv;

    if (scout_boot_flags & SCOUT_BOOT_F_VERBOSE) {
	dump_config();
    } /* if */

    xTraceInit();
    initRom();
#if 0
    addRomFlags();	/* should we support boot cmd line rom options? */
#endif
    xTimeInit();
    map_init();
    msgInit();

    ticks_per_sec = scout_ticks_per_second / SCOUT_TICKS_PER_CLOCK;
    evInit((1000000 + ticks_per_sec / 2) / ticks_per_sec);
    xShepherdInit();
    netMaskInit();
    prottbl_init();
    upiInit();
    build_pgraph_dev();
    build_pgraph();

    xTrace0(init, TR_EVENTS, "init thread exits");
} /* xkInit */

			/*** end of xinit.c ***/
