/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_locore.s,v $
 *
 * HISTORY
 * $Log: alpha_locore.s,v $
 * Revision 1.5  1994/06/11  00:02:39  menze
 * Added copyright notice
 *
 * Revision 1.4  1994/04/30  00:50:02  davidm
 * Clean up and SCOUT_TICKS_PER_CLOCK replaces hardcode constant 16.
 *
 * Revision 1.3  1994/04/27  05:14:35  davidm
 * (unaligned_fault): added.
 *
 * (interrupt): support for !SCOUT_ROLLFORWARD added.
 *
 * Revision 1.2  1994/04/21  07:18:16  davidm
 * Minor changes.
 *
 */
/*
 * Miscellaneous low-level assembly routines.
 */
#include <regdef.h>
#include "alpha_asm.h"
#include "alpha_pal.h"
#include "scout_config.h"
#include "scout_sync_arch.h"

#ifdef TESTING
	
	.extern	current_ipl
	
# define RTI \
	ldq	a0, PF_PS(sp); \
	lda	a1, current_ipl; \
	stq	a0, (a1); \
	ldq	gp, PF_GP(sp); \
	ldq	fp, PF_PC(sp); /* bad things happen if fp is used! */ \
	ldq	a0, PF_A0(sp); \
	ldq	a1, PF_A1(sp); \
	ldq	a2, PF_A2(sp); \
	lda	sp, PF_SIZE(sp); \
	jmp	(fp)
#else
# define RTI	call_pal PAL_rti
#endif

	.set noat

	.extern alpha_interrupt_handler
	.extern alpha_instruction_fault
	.extern alpha_mm_fault

	.globl interrupt

	.text

.ent interrupt 0	# nesting level 0
	.frame sp IF_SIZE ra # sp is frame pointer, 0xd0 sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
interrupt:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)
	stq	 s0, IF_S0(sp)

#ifndef SCOUT_ROLLFORWARD
	/*
	 * Deal with hard clock interrupts quickly:
	 */
	cmpeq	a0, 1, v0		# is it a clock interrupt?
	blbc	v0, full_interrupt	# no ->

	lda	v0, scout_age		# yup, increment scout_age
	ldq	s0, (v0)
	addq	s0, 1, s0
	stq	s0, (v0)
	and	s0, SCOUT_TICKS_PER_CLOCK-1, v0	# (scout_age % 16) == 0?
	beq	v0, play_leds		# yes -> need to do some more work
	ldq	v0, IF_V0(sp)
	ldq	s0, IF_S0(sp)
	lda	sp, IF_SIZE(sp)
	RTI

play_leds:
#ifdef SCOUT_PLAY_LEDS
# if 0
	/* reset system diagnostic LEDs on the back of the system: */
	lda	v0, 0xfffffc0000000000 | SSR
	ldl	s0, (v0)
	andnot	s0, SSR_DIAG_LED_MASK, s0
	stl	s0, (v0)
# else
	/*
	 * Reset bottom right 8 pixels on a PMAG-BA graphics board in
	 * TURBOchannel slot #3 (aka 0, on the DEC 3000/600):
	 */
	and	s0, 0x1f, v0
	bne	v0, full_interrupt
	lda	v0, 0xfffffc01603400f8
	stq	zero, (v0)
# endif
#endif /* SCOUT_PLAY_LEDS */

full_interrupt:
#endif /* SCOUT_ROLLFORWARD */

	stq	 s1, IF_S1(sp)
	stq	 s2, IF_S2(sp)
	stq	 s3, IF_S3(sp)
	stq	 s4, IF_S4(sp)
	stq	 s5, IF_S5(sp)
	stq	 s6, IF_S6(sp)

	stq	 a3, IF_A3(sp)
	stq	 a4, IF_A4(sp)
	stq	 a5, IF_A5(sp)
	stq	 t0, IF_T0(sp)
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 t4, IF_T4(sp)
	stq	 t5, IF_T5(sp)
	stq	 t6, IF_T6(sp)
	stq	 t7, IF_T7(sp)
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	 ra, IF_RA(sp)
	stq	t12, IF_T12(sp)
	stq	$at, IF_AT(sp)

	lda	pv, alpha_interrupt_handler
	jsr	ra, (pv), alpha_interrupt_handler

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	 ra, IF_RA(sp)
	ldq	t12, IF_T12(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	RTI
.end interrupt

	.globl	instruction_fault

.ent instruction_fault 0	# nesting level 0
	.frame sp IF_SIZE ra # sp is frame pointer, 0xd0 sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
instruction_fault:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)
	stq	 s0, IF_S0(sp)
	stq	 s1, IF_S1(sp)
	stq	 s2, IF_S2(sp)
	stq	 s3, IF_S3(sp)
	stq	 s4, IF_S4(sp)
	stq	 s5, IF_S5(sp)
	stq	 s6, IF_S6(sp)

	stq	 a3, IF_A3(sp)
	stq	 a4, IF_A4(sp)
	stq	 a5, IF_A5(sp)
	stq	 t0, IF_T0(sp)
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 t4, IF_T4(sp)
	stq	 t5, IF_T5(sp)
	stq	 t6, IF_T6(sp)
	stq	 t7, IF_T7(sp)
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	 ra, IF_RA(sp)
	stq	t12, IF_T12(sp)
	stq	$at, IF_AT(sp)

	mov	sp, a1

	lda	pv, alpha_instruction_fault
	jsr	ra, (pv), alpha_instruction_fault

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	 ra, IF_RA(sp)
	ldq	t12, IF_T12(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	RTI
.end instruction_fault

	.globl	mm_fault

.ent mm_fault 0	# nesting level 0
	.frame sp IF_SIZE ra # sp is frame pointer, 0xd0 sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
mm_fault:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)
	stq	 s0, IF_S0(sp)
	stq	 s1, IF_S1(sp)
	stq	 s2, IF_S2(sp)
	stq	 s3, IF_S3(sp)
	stq	 s4, IF_S4(sp)
	stq	 s5, IF_S5(sp)
	stq	 s6, IF_S6(sp)

	stq	 a3, IF_A3(sp)
	stq	 a4, IF_A4(sp)
	stq	 a5, IF_A5(sp)
	stq	 t0, IF_T0(sp)
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 t4, IF_T4(sp)
	stq	 t5, IF_T5(sp)
	stq	 t6, IF_T6(sp)
	stq	 t7, IF_T7(sp)
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	 ra, IF_RA(sp)
	stq	t12, IF_T12(sp)
	stq	$at, IF_AT(sp)

	mov	sp, a3

	lda	pv, alpha_mm_fault
	jsr	ra, (pv), alpha_mm_fault

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	 ra, IF_RA(sp)
	ldq	t12, IF_T12(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	RTI
.end mm_fault

	.globl	unaligned_fault

.ent unaligned_fault 0	# nesting level 0
	.frame sp IF_SIZE ra # sp is frame pointer, 0xd0 sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
unaligned_fault:
	lda	sp, -IF_SIZE(sp)

	.prologue 1		# end of prologue, yes we use gp

	stq	 v0, IF_V0(sp)
	stq	 s0, IF_S0(sp)
	stq	 s1, IF_S1(sp)
	stq	 s2, IF_S2(sp)
	stq	 s3, IF_S3(sp)
	stq	 s4, IF_S4(sp)
	stq	 s5, IF_S5(sp)
	stq	 s6, IF_S6(sp)

	stq	 a3, IF_A3(sp)
	stq	 a4, IF_A4(sp)
	stq	 a5, IF_A5(sp)
	stq	 t0, IF_T0(sp)
	stq	 t1, IF_T1(sp)
	stq	 t2, IF_T2(sp)
	stq	 t3, IF_T3(sp)
	stq	 t4, IF_T4(sp)
	stq	 t5, IF_T5(sp)
	stq	 t6, IF_T6(sp)
	stq	 t7, IF_T7(sp)
	stq	 t8, IF_T8(sp)
	stq	 t9, IF_T9(sp)
	stq	t10, IF_T10(sp)
	stq	t11, IF_T11(sp)
	stq	 ra, IF_RA(sp)
	stq	t12, IF_T12(sp)
	stq	$at, IF_AT(sp)

	mov	sp, a3

	lda	pv, alpha_unaligned_fault
	jsr	ra, (pv), alpha_unaligned_fault

	ldq	 v0, IF_V0(sp)
	ldq	 t0, IF_T0(sp)
	ldq	 t1, IF_T1(sp)
	ldq	 t2, IF_T2(sp)
	ldq	 t3, IF_T3(sp)
	ldq	 t4, IF_T4(sp)
	ldq	 t5, IF_T5(sp)
	ldq	 t6, IF_T6(sp)
	ldq	 t7, IF_T7(sp)
	ldq	 s0, IF_S0(sp)
	ldq	 s1, IF_S1(sp)
	ldq	 s2, IF_S2(sp)
	ldq	 s3, IF_S3(sp)
	ldq	 s4, IF_S4(sp)
	ldq	 s5, IF_S5(sp)
	ldq	 s6, IF_S6(sp)

	ldq	 a3, IF_A3(sp)
	ldq	 a4, IF_A4(sp)
	ldq	 a5, IF_A5(sp)
	ldq	 t8, IF_T8(sp)
	ldq	 t9, IF_T9(sp)
	ldq	t10, IF_T10(sp)
	ldq	t11, IF_T11(sp)
	ldq	 ra, IF_RA(sp)
	ldq	t12, IF_T12(sp)
	ldq	$at, IF_AT(sp)

	lda	sp, IF_SIZE(sp)
	RTI
.end unaligned_fault

	#
	# This is a helper for the KGDB stub.  It expets the following
	# arguments:
	#
	#	a0 = pointer to saved state (Alpha_Saved_State)
	#	a1 = new stack pointer
	#
	.globl	kgdb_restore_and_rti

.ent kgdb_restore_and_rti 0	# nesting level 0
	.frame sp IF_SIZE ra # sp is frame pointer, 0xd0 sized stack
				# frame, ra is return address (that's a lie)
	.mask 1 | (1<<26), 0	# reserve space for all 31 regs @ offset 0
	.fmask 0, 0		# no FP regs saved
kgdb_restore_and_rti:
	.prologue 1		# end of prologue, yes we use gp

	#
	# Copy PAL frame from saved state to stack so we can
	# rti to it:
	#
	ldq	 t0, PF_PS+IF_SIZE(a0)
	ldq	 t1, PF_PC+IF_SIZE(a0)
	ldq	 t2, PF_GP+IF_SIZE(a0)
	ldq	 t3, PF_A0+IF_SIZE(a0)
	ldq	 t4, PF_A1+IF_SIZE(a0)
	ldq	 t5, PF_A2+IF_SIZE(a0)

	lda	 a1, -PF_SIZE(a1)	# alloc PAL frame on stack

	stq	 t0, PF_PS(a1)
	stq	 t1, PF_PC(a1)
	stq	 t2, PF_GP(a1)
	stq	 t3, PF_A0(a1)
	stq	 t4, PF_A1(a1)
	stq	 t5, PF_A2(a1)
	
	ldq	 v0, IF_V0(a0)
	ldq	 t0, IF_T0(a0)
	ldq	 t1, IF_T1(a0)
	ldq	 t2, IF_T2(a0)
	ldq	 t3, IF_T3(a0)
	ldq	 t4, IF_T4(a0)
	ldq	 t5, IF_T5(a0)
	ldq	 t6, IF_T6(a0)
	ldq	 t7, IF_T7(a0)
	ldq	 s0, IF_S0(a0)
	ldq	 s1, IF_S1(a0)
	ldq	 s2, IF_S2(a0)
	ldq	 s3, IF_S3(a0)
	ldq	 s4, IF_S4(a0)
	ldq	 s5, IF_S5(a0)
	ldq	 s6, IF_S6(a0)

	ldq	 a3, IF_A3(a0)
	ldq	 a4, IF_A4(a0)
	ldq	 a5, IF_A5(a0)
	ldq	 t8, IF_T8(a0)
	ldq	 t9, IF_T9(a0)
	ldq	t10, IF_T10(a0)
	ldq	t11, IF_T11(a0)
	ldq	 ra, IF_RA(a0)
	ldq	t12, IF_T12(a0)
	ldq	$at, IF_AT(a0)

	mov	a1, sp		# switch to new stack

	RTI
.end kgdb_restore_and_rti
